<?php

namespace Drupal\html_segmenter;

use Drupal\html_segmenter\HtmlSegmenterInterface;
use Drupal\Component\Utility\Html;

/**
 * Transforms translatable field values into MT translatable XML strings.
 */
class HtmlSegmenter implements HtmlSegmenterInterface {

    /**
     * Regex patterns for non-translatable Drupal variables in UI strings.
     *
     * @var array
     */
    private static $drupalVariablePatterns = array(
        // Match @.. and %.. and !..
        '(([^(@|%|!)^\sa-zA-Z\d+<>]*)(@|%|!)([^\s<>(@|%|!)]+))',
        // Match curly bracket variables like {var}.
        '({([^\s]*?)})',
        // Match double curly bracket variables like {{ var }}.
        '({{\s*(.*?)\s*}})',
        // Match square bracket variables like [site:name].
        '(\[(.*?)\:(.*?)\])',
        // Match PHP variables like $var['abc'].
        '(\$(([a-zA-Z_\x7f-\xff][a-zA-Z0-9_\x7f-\xff]*(->)*)*(\[[^\]]*\])*))',
    );

    /**
     * Char dictionary for chars to replace before string is converted to HTML.
     *
     * @var array
     */
    private static $specialCharDict = array("\x03" => '<etx></etx>');
    /**
     * List of translatable HTML attributes.
     *
     * @var array
     */
    private static $defaultTranslatableHtmlAttributes = [
        'alt',
        'title',
        'aria-label',
        'media-alt',
        'media-title',
    ];

    /**
     * The actual list of translatable HTML attributes for this instance.
     *
     * @var array
     */
    private $translatableHtmlAttributes;

    /**
   * {@inheritdoc}
   */
    public function __construct(array $translatableHtmlAttributes = []) {
        $this->translatableHtmlAttributes = $translatableHtmlAttributes ?: self::$defaultTranslatableHtmlAttributes;
        if (class_exists('\Drupal') && method_exists('\Drupal', 'moduleHandler') && \Drupal::hasContainer()) {
            \Drupal::moduleHandler()->alter('html_segmenter_translatable_attributes', $this->translatableHtmlAttributes);
        }
    }

    /**
     * Converts strings to HTML and extracts translatable text and attribute values
     *
     * @param string[] $values
     *   Translatable string values, that can be HTML.
     *
     * @return string[]
     *   Translatable strings
     */
    public function extractTranslatableHtmlValues(array $values): array {
        $texts_to_translate = array();
        foreach ($values as $v) {
            if (!is_string($v)) {
                throw new \InvalidArgumentException('All values must be strings.');
            }
            $html_string = self::wrapAsHtml($v);
            $texts = $this->extractTextsFromHtml($html_string);
            $texts_to_translate = array_merge($texts_to_translate, $texts);
        }
        return array_values(array_filter($texts_to_translate, function($string) {
            return !self::isStringEmptyOrBlankSpace($string);
        }));
    }

    /**
     * Merges raw translation list to a list of translations for original strings
     *
     * @param string[] $original_strings Original values.
     * @param string[] $extracted_values Extracted HTML translatable values.
     * @param string[] $raw_translations Raw translations from MT service.
     * @return string[] Translations for original values.
     */
    public function mergeTranslatedHtmlValues($original_strings, $extracted_values, $raw_translations): array {
        if (count($extracted_values) !== count($raw_translations)) {
            throw new \InvalidArgumentException('The number of extracted values and translations must match.');
        }
        $translations = self::restoreSpaces($extracted_values, $raw_translations);
        $translations = $this->replaceStringTranslations($original_strings, $translations);
        return $translations;
    }

    /**
     * Goes through HTML DOM and extracts translatable text content and attribute values.
     *
     * @param string $html_content HTML string.
     * @return string[] Translatable strings.
     */
    private function extractTextsFromHtml($html_content): array {
        try {
            $dom = Html::load($html_content);
        } catch (\Exception $e) {
            throw new \RuntimeException('Invalid HTML content provided: ' . $e->getMessage());
        }

        $texts_to_translate = array();
        $html_content = str_replace(array_keys(self::$specialCharDict), self::$specialCharDict, $html_content);
        // Extract text content from all text nodes.
        $this->traverseDom($dom, function($node, $attr = null) use (&$texts_to_translate) {
            if ($attr) {
                $texts_to_translate[] = $node->getAttribute($attr);
            } else {
                $translatable_parts = self::splitNodeTranslatableParts($node->nodeValue)[0];
                $texts_to_translate = array_merge($texts_to_translate, $translatable_parts);
            }
        });
        return $texts_to_translate;
    }

    /**
     * Inserts translated node & attribute values into original strings.
     *
     * @param string[] $source_values
     *   Original strings that were translated.
     * @param string[] $translated_values
     *   String value translations.
     *
     * @return string[]
     *   Translated strings
     */
    private function replaceStringTranslations($original_values, $translated_values): array {
        $index = 0;
        $translations = array();
        foreach ($original_values as $v) {
            $html_content = self::wrapAsHtml($v);
            $html_content = str_replace(array_keys(self::$specialCharDict), self::$specialCharDict, $html_content);
            $dom = Html::load($html_content);

            $this->traverseDom($dom, function($node, $attr = null) use (&$translated_values, &$index) {
                if ($attr && !self::isStringEmptyOrBlankSpace($node->getAttribute($attr))) {
                    $node->setAttribute($attr, $translated_values[$index++]);
                }
                else {
                    $parts = self::splitNodeTranslatableParts($node->nodeValue);
                    $translatable_count = count($parts[0]);
                    for ($i = 0; $i < $translatable_count; $i++) {
                        if (self::isStringEmptyOrBlankSpace($parts[0][$i])) {
                            continue;
                        }
                        $parts[0][$i] = $translated_values[$index++]; // replace translatable parts with their translations.
                    }
                    $node->nodeValue = self::nodePartsToString($parts);
                }
            });

            $wrapper = $dom->documentElement->lastChild->lastChild;
            $translation = self::getInnerHtml($wrapper);
            $translation = str_replace(self::$specialCharDict, array_keys(self::$specialCharDict), $translation);
            $translations[] = $translation;
        }
        return $translations;
    }

    private static function isStringEmptyOrBlankSpace($string) {
        return empty(trim($string));
    }

    /**
     * Traverses each HTML element in DOM and executes callback when text or attribute value is found.
     *
     * @param DOMDocument $dom HTML DOM.
     * @param callable    $callback Callback function.
     * @return void
     */
    private function traverseDom($dom, $callback): void {
        $text_nodes = $dom->getElementsByTagName('*');
        foreach ($text_nodes as $node) {
            foreach ($node->childNodes as $child) {
                if ($child->nodeType === XML_TEXT_NODE && trim($child->nodeValue) !== '') {
                    $callback($child);
                }
            }
            foreach ($this->translatableHtmlAttributes as $attr) {
                if ($node->hasAttribute($attr)) {
                    $callback($node, $attr);
                }
            }
        }
    }

    /**
     * Restores spaces around HTML tags removed by MT service.
     *
     * @param string[] $original_values Original values.
     * @param string[] $translated_values Translated values.
     *
     * @return string[]
     *   Translations
     */
    private static function restoreSpaces($original_values, $translated_values): array {
        // Fix leading, trailing & double spaces.
        $count = count($original_values);
        for ($i = 0; $i < $count; $i++) {
            // Replace different types of UTF-8 spaces (incl. nbsp) with regular space.
            $original = preg_replace('%\s+%u', ' ', $original_values[$i]);
            $translation = preg_replace('%\s+%u', ' ', $translated_values[$i]);

            if (ctype_space(substr($original, 0, 1)) && ! ctype_space(substr($translation, 0, 1))) {
                $translated_values[$i] = substr($original, 0, 1) . $translated_values[$i];
            }
            if (ctype_space(substr($original, -1)) && ! ctype_space(substr($translation, -1))) {
                $translated_values[$i] = $translated_values[$i] . substr($original, -1);
            }
            $translated_values[$i] = str_replace('  ', ' ', $translated_values[$i]);
        }
        return $translated_values;
    }

    /**
     * Get inner HTML content of a node.
     *
     * @param \DOMNode $node HTML node.
     * @return string HTML content as a string.
     */
    private static function getInnerHtml($node) {
        $innerHTML = '';
        $children = $node->childNodes;
        foreach ($children as $child) {
            $innerHTML .= $child->ownerDocument->saveXML($child);
        }
        return $innerHTML;
    }

    /**
     * Wrap translatable string in HTML tags.
     *
     * @param string $string
     *   Source string.
     *
     * @return string
     *   HTML node string
     */
    private static function wrapAsHtml($string) {
        return '<div>' . $string . '</div>';
    }

    /**
     * Splits translatable and non-translatable parts of a node.
     *
     * @param string $string Source string.
     * @return array Strings of node grouped by translatable/non-translatable status.
     */
    private static function splitNodeTranslatableParts($string): array {
        $non_translatables_pattern = '/' . implode('|', self::$drupalVariablePatterns) . '/';
        preg_match_all($non_translatables_pattern, $string, $matches, PREG_OFFSET_CAPTURE);
        $matches = $matches[0];

        $translatable_parts = array();
        $non_translatable_parts = array(''); // empty value helps to restore string afterwards by ensuring it always starts with non-translatable part.

        if (empty($matches)) {
            if ( ! empty( $string ) ) {
                $translatable_parts[] = $string;
            }
        }
        else {
            $previous_index = 0;

            foreach ($matches as $match) {
                $match_position = $match[1];
                if ($match_position === 0) {
                    $non_translatable_parts[0] = $match[0]; // overwrite empty value if string actually starts with non-translatable part.
                } else {
                    $non_translatable_parts[] = $match[0];
                }
                if ($match_position > $previous_index) {
                    $translatable_parts[] = substr($string, $previous_index, $match_position - $previous_index);
                }
                $previous_index = $match_position + strlen($match[0]);
            }

            if ($previous_index < strlen($string)) {
                $translatable_parts[] = substr($string, $previous_index);
            }
        }
        return array($translatable_parts, $non_translatable_parts);
    }

    /**
     * Combines node translatable and non translatable parts back into string
     *
     * @param array $parts Array of translatable and non-translatable string arrays.
     * @return string Node value.
     */
    private static function nodePartsToString($parts): string {
        $combined = array();
        $translatable_parts = $parts[0];
        $non_translatable_parts = $parts[1];
        $count = max(count($translatable_parts), count($non_translatable_parts));
        for ($i = 0; $i < $count; $i++) {
            if (isset($non_translatable_parts[$i])) {
                $combined[] = $non_translatable_parts[$i];
            }
            if (isset($translatable_parts[$i])) {
                $combined[] = $translatable_parts[$i];
            }
        }
        return implode('', $combined);
    }
}